@@ -1,13 +1,13 @@
 from __future__ import absolute_import
 
+import os
 import socket
 import errno
 import logging
-from itertools import chain
 from contextlib import contextmanager
 from abc import ABCMeta, abstractproperty
 
-from pyuv import Pipe, TCP, Poll, UV_READABLE
+from pyuv import TCP, Poll, UV_READABLE
 from pyuv.errno import strerror
 from six import with_metaclass
 
@@ -15,11 +15,13 @@ from thriftworker.constants import BACKLOG_SIZE
 from thriftworker.utils.mixin import LoopMixin, StartStopMixin
 from thriftworker.utils.loop import in_loop
 from thriftworker.utils.decorators import cached_property
-from thriftworker.utils.waiting import wait
+from thriftworker.utils.waiter import Waiter
+
+from . import utils
 
 logger = logging.getLogger(__name__)
 
-Handle = TCP if hasattr(TCP, 'open') else Pipe
+NOTBLOCK = (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINVAL, errno.EBADF)
 
 
 @contextmanager
@@ -27,17 +29,18 @@ def ignore_eagain():
     """Ignore all *EAGAIN* errors in context."""
     try:
         yield
-    except socket.error as exc:
+    except OSError as exc:
         if exc.errno not in (errno.EAGAIN, errno.EWOULDBLOCK,
                              errno.EINVAL, errno.EBADF):
             raise
 
 
 class Connections(object):
-    """Store connections."""
+    """Store existed connections."""
 
     def __init__(self):
         self.connections = set()
+        self._callback = None
 
     def __len__(self):
         return len(self.connections)
@@ -48,6 +51,24 @@ class Connections(object):
     def __repr__(self):
         return repr(self.connections)
 
+    def _execute_callback(self):
+        """Execute callback if needed."""
+        if self._callback is not None:
+            self._callback()
+            self._callback = None
+
+    @property
+    def callback(self):
+        """Return callback if existed."""
+        return self._callback
+
+    @callback.setter
+    def callback(self, cb):
+        """Setup callback and execute it if needed."""
+        self._callback = cb
+        if not self.connections:
+            self._execute_callback()
+
     def register(self, connection):
         """Register new connection."""
         self.connections.add(connection)
@@ -58,6 +79,8 @@ class Connections(object):
             self.connections.remove(connection)
         except KeyError:
             pass
+        if not self.connections:
+            self._execute_callback()
 
     def close(self):
         connections = self.connections
@@ -66,6 +89,7 @@ class Connections(object):
             if not connection.is_closed():
                 logger.warn('Connection %r closed prematurely', connection)
                 connection.close()
+        self._execute_callback()
 
 
 class BaseAcceptor(with_metaclass(ABCMeta, LoopMixin)):
@@ -87,9 +111,7 @@ class BaseAcceptor(with_metaclass(ABCMeta, LoopMixin)):
     @cached_property
     def _socket(self):
         """Create socket from given descriptor."""
-        socket = self.app.env.socket
-        sock = socket.fromfd(self.descriptor, socket.AF_INET,
-                             socket.SOCK_STREAM)
+        sock = socket.fromfd(self.descriptor, socket.AF_INET, socket.SOCK_STREAM)
         sock.setblocking(0)
         return sock
 
@@ -106,6 +128,11 @@ class BaseAcceptor(with_metaclass(ABCMeta, LoopMixin)):
         """Return number of active connections."""
         return len(self._connections)
 
+    @property
+    def empty(self):
+        """Is acceptor empty or not."""
+        return self.connections_number == 0
+
     @property
     def active(self):
         """Is current acceptor active."""
@@ -121,10 +148,9 @@ class BaseAcceptor(with_metaclass(ABCMeta, LoopMixin)):
         loop = self.loop
         service = self.name
         connections = self._connections
-        listen_sock = self._socket
+        listen_fd = self._socket.fileno()
         worker = self.app.worker
         producer = worker.create_producer(service)
-        concurrency = worker.concurrency
 
         def on_close(connection):
             """Callback called when connection closed."""
@@ -136,17 +162,23 @@ class BaseAcceptor(with_metaclass(ABCMeta, LoopMixin)):
                 logger.error('Error handling new connection for'
                              ' service %r: %s', service, strerror(error))
                 return
-            if concurrency.reached:
+            try:
+                fd, addr = utils.accept_connection(listen_fd)
+            except OSError as exc:
+                if exc.errno not in NOTBLOCK:
+                    raise
                 return
-            with ignore_eagain():
-                sock, addr = listen_sock.accept()
-                # Disable Nagle's algorithm for socket.
-                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
-                client = Handle(loop)
-                client.open(sock.fileno())
-                connection = self.Connection(producer, loop, client, sock,
-                                             on_close)
-                connections.register(connection)
+            try:
+                # Setup socket.
+                utils.set_nonblocking(fd)
+                utils.set_sockopt(fd, socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+            except:
+                os.close(fd)
+                raise
+            handle = TCP(loop)
+            handle.open(fd)
+            connection = self.Connection(producer, loop, handle, addr, on_close)
+            connections.register(connection)
 
         return inner_acceptor
 
@@ -158,11 +190,12 @@ class BaseAcceptor(with_metaclass(ABCMeta, LoopMixin)):
             poller.start(UV_READABLE, self.acceptor)
 
     @in_loop
-    def stop(self):
+    def stop(self, callback=None):
         """Stop acceptor if active."""
         poller = self._poller
         if poller.active and not poller.closed:
             poller.stop()
+        self._connections.callback = callback
 
     @in_loop
     def close(self):
@@ -178,6 +211,8 @@ class Acceptors(StartStopMixin, LoopMixin):
 
     def __init__(self):
         self._acceptors = {}
+        self._stop_waiter = Waiter(
+            timeout=self.app.shutdown_timeout)
         super(Acceptors, self).__init__()
 
     def __iter__(self):
@@ -208,24 +243,35 @@ class Acceptors(StartStopMixin, LoopMixin):
         for acceptor in self._acceptors.values():
             acceptor.start()
 
-    def stop_accepting(self):
+    def stop_accepting(self, callback=None):
         """Stop all registered acceptors if needed."""
         for acceptor in self._acceptors.values():
-            acceptor.stop()
+            acceptor.stop(callback)
 
     @property
     def connections_number(self):
+        """Return current connection number across all acceptors."""
         return sum(acceptor.connections_number for acceptor in self)
 
+    @property
+    def empty(self):
+        """Are all acceptors empty or not."""
+        return self.connections_number == 0
+
     def stop(self):
         """Close all registered acceptors."""
-        self.stop_accepting()
         # wait for unclosed connections
-        if self.connections_number:
-            for connection in chain.from_iterable(self):
-                logger.info('Wait for %r...', connection)
-            if wait(lambda: self.connections_number == 0,
-                    timeout=self.app.shutdown_timeout):
-                logger.debug('All connections closed...')
+        def on_close():
+            if self.empty:
+                self._stop_waiter.done()
+        # stop accepting new connection
+        self.stop_accepting(callback=on_close)
+        # wait for unclosed connections
+        if not self.empty:
+            logger.info('Waiting for unclosed connections...')
+        self._stop_waiter.wait()
+        if not self.empty:
+            logger.warning('Not all connection closed!')
+        # close existed connection
         for acceptor in self:
             acceptor.close()
