@@ -14,10 +14,14 @@ def rgetattr(obj, path):
     return reduce(getattr, [obj] + path.split('.'))
 
 
+def djb2(L):
+    return reduce(lambda h,c: ord(c) + ((h << 5) + h), L, 5381)
+
+
 def get_port_from_range(name, range_start, range_end):
     """Get random port from given range [range_start, range_end]."""
     # Detect port for this service.
-    hashed_port = abs(hash(name) & 0xffff)
+    hashed_port = djb2(name) & 0xffff
     # Ensure that detected port is in pool.
     while hashed_port > range_start:
         hashed_port = hashed_port - range_start
