@@ -8,7 +8,7 @@ from functools import partial
 from six import with_metaclass
 
 from ..utils.mixin import LoopMixin, StartStopMixin
-from ..utils.atomics import ContextCounter
+from ..utils.atomics import AtomicBoolean, AtomicInteger
 from ..utils.decorators import cached_property
 from ..utils.monotime import monotonic
 
@@ -78,27 +78,33 @@ class BaseWorker(StartStopMixin, with_metaclass(ABCMeta, LoopMixin)):
 
     def __init__(self, pool_size=None):
         self.pool_size = pool_size or 10
+        self.last_overflow = None
+        self.accepting = AtomicBoolean(True)
+        self.queued = AtomicInteger(0)
         super(BaseWorker, self).__init__()
 
     def create_callback(self):
         """Create callback that should be called after request was done."""
-        concurrency = self.concurrency
+        accepting = self.accepting
+        queued = self.queued
         pool_size = self.pool_size
         acceptors = self.app.acceptors
         counter = self.app.counters['response_served']
         timeouts = self.app.timeouts
         execution_timers = self.app.execution_timers
         dispatching_timers = self.app.dispatching_timers
-        delay = self.app.hub.callback
+        call_soon = self.app.hub.call_soon
+        loop = self.app.loop
 
         def start_accepting():
-            if not concurrency.reached:
+            if accepting:
                 return
-            concurrency.reached.clean()
-            logger.info('Start registered acceptors,'
-                        ' current concurrency: %d...', int(concurrency))
+            accepting.set()
             acceptors.start_accepting()
 
+            logger.info('Start acceptors, queued %d tasks...',
+                        int(queued))
+
         def inner_callback(request, result, exception=None):
             """Process task result."""
             method_name = request.method_name
@@ -118,8 +124,8 @@ class BaseWorker(StartStopMixin, with_metaclass(ABCMeta, LoopMixin)):
                 execution_timers[method_name] += request.execution_time
                 dispatching_timers[method_name] += request.dispatching_timers
 
-            if concurrency.reached and pool_size > concurrency:
-                delay(start_accepting)
+            if not accepting and queued == 0:
+                call_soon(start_accepting)
 
         return inner_callback
 
@@ -127,25 +133,21 @@ class BaseWorker(StartStopMixin, with_metaclass(ABCMeta, LoopMixin)):
     def create_consumer(self):
         raise NotImplementedError()
 
-    @cached_property
-    def concurrency(self):
-        """How many tasks executed in parallel?"""
-        return ContextCounter()
-
     def create_task(self, processor):
         """Create new task for given processor."""
-        concurrency = self.concurrency
+        queued = self.queued
 
         def inner_task(request):
             """Process incoming request with given processor."""
-            with concurrency:
-                return request.execute(processor)
+            queued.decr()
+            return request.execute(processor)
 
         return inner_task
 
     def create_producer(self, service):
         """Create producer for connections."""
-        concurrency = self.concurrency
+        accepting = self.accepting
+        queued = self.queued
         pool_size = self.pool_size
         callback = self.create_callback()
         processor = self.app.services.create_processor(service)
@@ -154,18 +156,20 @@ class BaseWorker(StartStopMixin, with_metaclass(ABCMeta, LoopMixin)):
         consume = self.create_consumer()
         acceptors = self.app.acceptors
         loop = self.app.loop
-        delay = self.app.hub.callback
+        call_soon = self.app.hub.call_soon
         Request = self.Request
 
         def stop_accepting():
-            if concurrency.reached or pool_size > concurrency:
+            if not accepting:
                 return
-            logger.info('Stop registered acceptors,'
-                        ' current concurrency: %d...', int(concurrency))
-            counter.add()
-            concurrency.reached.set()
+            accepting.clean()
             acceptors.stop_accepting()
 
+            self.last_overflow = loop.now()
+            counter.add()
+            logger.info('Stop acceptors, queued %d tasks...',
+                        int(queued))
+
         def inner_producer(connection, message_buffer, request_id):
             """Enqueue given request to thread pool."""
             request = Request(loop=loop,
@@ -175,7 +179,9 @@ class BaseWorker(StartStopMixin, with_metaclass(ABCMeta, LoopMixin)):
                               service=service)
             curried_task = partial(task, request)
             consume(curried_task, partial(callback, request))
-            if not concurrency.reached and pool_size <= concurrency:
-                delay(stop_accepting)
+            queued.incr()
+            if accepting and queued > pool_size:
+                call_soon(stop_accepting)
 
         return inner_producer
+
